# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: ssc-sensor-suid.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'ssc-sensor-suid.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import ssc_common_pb2 as ssc__common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15ssc-sensor-suid.proto\x1a\x10ssc-common.proto\"X\n\x0eSscSuidRequest\x12\x11\n\tdata_type\x18\x01 \x02(\t\x12\x16\n\x0e\x65nable_updates\x18\x02 \x01(\x08\x12\x1b\n\x13only_default_values\x18\x03 \x01(\x08\":\n\x0fSscSuidResponse\x12\x11\n\tdata_type\x18\x01 \x02(\t\x12\x14\n\x03uid\x18\x02 \x03(\x0b\x32\x07.SscUid')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ssc_sensor_suid_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SSCSUIDREQUEST']._serialized_start=43
  _globals['_SSCSUIDREQUEST']._serialized_end=131
  _globals['_SSCSUIDRESPONSE']._serialized_start=133
  _globals['_SSCSUIDRESPONSE']._serialized_end=191
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python3
#
# libssc: Library to expose Qualcomm Sensor Core sensors
# Copyright (C) 2022-2026 Dylan Van Assche
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import sys
from enum import Enum
from typing import Tuple


class ValueType(Enum):
    INT = 0
    ARRAY = 1
    SEQUENCE = 2


class MessageType(Enum):
    REQUEST = 0
    RESPONSE = 2
    INDICATION = 4
    UNKNOWN = sys.maxsize


class QMI():
    """QMI messages encoding and decoding functions."""

    @staticmethod
    def parse_header(message: bytes) -> Tuple[MessageType, int, int, bytes]:
        """
        Parse a QMI header to extract:
          - Type of message (uint8)
          - Transaction ID (uint16)
          - Message ID (uint16)
          - Message length (uint16)

        Parameters
        ----------
        message : bytes
            Raw QMI message as bytes (little endian).

        Returns
        -------
        message_type : MessageType
            Type of message e.g., REQUEST, RESPONSE, or INDICATION.
        transaction_id : int
            ID of the transaction for concurrent messages.
        message_id : int
            ID of the message, specific for the QMI service.
        message_length : int
            Length of the message in bytes.
        """

        # QMI message type
        if message[0] == MessageType.REQUEST.value:
            message_type = MessageType.REQUEST
        elif message[0] == MessageType.RESPONSE.value:
            message_type = MessageType.RESPONSE
        elif message[0] == MessageType.INDICATION.value:
            message_type = MessageType.INDICATION
        else:
            print(f'Unknown message type: {message[0]}', file=sys.stderr)
            message_type = MessageType.UNKNOWN 

        # QMI transaction ID
        transaction_id = int.from_bytes(message[1:3], byteorder='little')

        # QMI message ID
        message_id = int.from_bytes(message[3:5], byteorder='little')

        # QMI message length
        message_length = int.from_bytes(message[5:7], byteorder='little')

        return message_type, transaction_id, message_id, message_length

    @staticmethod
    def generate_header(message_type: MessageType, transaction_id: int, message_id: int) -> bytes:
        """
        Generate a QMI message with no TLVs and a QMI header.

        Parameters
        ----------
        message_type : MessageType
            Type of QMI message
        transaction_id : int
            ID of the transaction.
        message_id : int
            ID of the message for the specific QMI service.

        Returns
        -------
        message : bytes
            Raw QMI message as bytes (little endian).
        """

        message = bytes()
        message += message_type.value.to_bytes(1, byteorder='little')
        message += transaction_id.to_bytes(2, byteorder='little')
        message += message_id.to_bytes(2, byteorder='little')
        message += (len(message) + 2).to_bytes(2, byteorder='little')

        return message

    @staticmethod
    def parse_tlv(message: bytes, tlv: int, value_type: ValueType) -> Tuple[int, int, bytes]:
        """
        Parse QMI TLV in a QMI message to extract:
          - Type (uint8)
          - Length (uint16)
          - Value (bytes)

        Parameters
        ----------
        message : bytes
            Raw QMI message as bytes (little endian).

        Returns
        -------
        tlvs : dict
            Dictionary of TLVs extracted with Type as dictionary's key
            and Value as dictionary's value. Length is implied.
        """
        message_type, transaction_id, message_id, message_length = QMI.parse_header(message)

        # Strip QMI header
        tlvs = message[7:]

        # Iterate over all TLVs
        offset = 0
        tlv_id = 0
        while offset < len(tlvs):
            tlv_type = int.from_bytes(tlvs[0+offset:1+offset], byteorder='little')
            tlv_length = int.from_bytes(tlvs[1+offset:3+offset], byteorder='little')
            tlv_value = tlvs[3+offset:3+tlv_length+offset]
            offset += 1 + 2 + tlv_length
            tlv_id += 1

            # Strip array length from QMI TLV value
            if value_type == ValueType.INT or value_type == ValueType.SEQUENCE:
                tlv_value = int.from_bytes(tlv_value, byteorder='little')
            elif value_type == ValueType.ARRAY:
                tlv_value = tlv_value[2:]

            # Look for TLV in message, skip others
            if tlv != tlv_type:
                continue

            return tlv_type, tlv_length, tlv_value

    @staticmethod
    def generate_tlv(message: bytes, tlv_id: int, tlv_value: bytes, value_type: ValueType) -> bytes:
        """
        Generate a QMI TLV and add it to the message.

        Parameters
        ----------
        message : bytes
            Raw QMI message as bytes (little endian).
        tlv_id : int
            TLV Type.
        tlv_value : bytes
            TLV Value.
        value_type : ValueType
            Type of the TLV Value.

        Returns
        -------
        message : bytes
            Raw QMI message as bytes (little endian).
        """
        # Add TLV and calculate its length
        message += tlv_id.to_bytes(1, byteorder='little')

        if value_type == ValueType.ARRAY:
            message += (len(tlv_value) + 2).to_bytes(2, byteorder='little')
        elif value_type == ValueType.INT or value_type == ValueType.SEQUENCE:
            message += len(tlv_value).to_bytes(2, byteorder='little')
        else:
            print(f'Unknown value type: {value_type}', file=sys.stderr)

        if value_type == ValueType.ARRAY:
            value_array_length = len(tlv_value)
            message += value_array_length.to_bytes(2, byteorder='little')

        message += tlv_value

        # Update QMI header length
        qmi_length = int.from_bytes(message[5:7], byteorder='little')
        message = message[:5] + \
                  (len(message) - 7).to_bytes(2, byteorder='little') + \
                  message[7:]

        return message



#!/usr/bin/env python3
#
# libssc: Library to expose Qualcomm Sensor Core sensors
# Copyright (C) 2022-2026 Dylan Van Assche
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import os
import sys
import time
import json
import random
from enum import Enum
from typing import Tuple
from glob import glob
from ssc_server.qmi import QMI, ValueType, MessageType
import ssc_server.ssc_sensor_suid_pb2 as SensorSuid
import ssc_server.ssc_common_pb2 as SscCommon
import ssc_server.ssc_sensor_proximity_pb2 as SensorProximity
import ssc_server.ssc_sensor_accelerometer_pb2 as SensorAccelerometer
import ssc_server.ssc_sensor_magnetometer_pb2 as SensorMagnetometer
import ssc_server.ssc_sensor_light_pb2 as SensorLight
import ssc_server.ssc_sensor_rotationvector_pb2 as SensorCompass

SSC_SERVICE_ID = 0x190
SSC_CLIENT_ID = 1

SSC_CONTROL_MSG_ID = 0x0020
SSC_CONTROL_MSG_SUCCESS_ID = 0x2
SSC_CONTROL_MSG_REPORT_TYPE_ID = 0x10
SSC_CONTROL_MSG_CLIENT_ID = 0x10
SSC_CONTROL_MSG_DATA_ID = 0x1

SSC_REPORT_SMALL_MSG_ID = 0x0021
SSC_REPORT_SMALL_CLIENT_ID = 0x01
SSC_REPORT_SMALL_DATA = 0x02
SSC_REPORT_LARGE_MSG_ID = 0x0022
SSC_REPORT_LARGE_CLIENT_ID = 0x01
SSC_REPORT_LARGE_DATA = 0x02

SSC_SUID_SENSOR_UID_HIGH = 0xABABABABABABABAB
SSC_SUID_SENSOR_UID_LOW = 0xABABABABABABABAB

SSC_PROTOBUF_DISCOVERY_MSG_ID = 768
SSC_PROTOBUF_DISCOVER_MSG_ID = 512
SSC_PROTOBUF_GET_ATTRIBUTE_MSG_ID = 1
SSC_PROTOBUF_RESPONSE_GET_ATTRIBUTES_MSG_ID = 128
SSC_PROTOBUF_ENABLE_REPORT_MSG_ID = 514
SSC_PROTOBUF_DISABLE_REPORT_MSG_ID = 10
SSC_PROTOBUF_ENABLE_CONTINUOUS_MSG_ID = 513


class ReportType(Enum):
    """
        SSC report type enum.
    """
    SMALL = 0x0
    LARGE = 0x1
    UNKNOWN = sys.maxsize


class DataType(Enum):
    """
        SSC sensor data types.
    """
    PROXIMITY = 'proximity'
    ACCELEROMETER = 'accel'
    MAGNETOMETER = 'mag'
    LIGHT = 'ambient_light'
    ROTATIONVECTOR = 'rotv'


class SSC():
    """
    Qualcomm's Snapdragon Sensor Core QMI messages.
    """

    @staticmethod
    def _read_data() -> dict:
        """
        Read JSON data with sensor configuration and fake measurements.

        Returns
        -------
        data : dict
            Sensor configuration and fake measurements.
        """
        data = []

        for sensor in glob(os.path.join(os.path.dirname(__file__), 'data', 'sensor-*.json')):
            with open(sensor) as f:
                data.append(json.load(f))

        print('SSC DATA', data)

        return data

    @staticmethod
    def parse_message_control_input(message: bytes) -> Tuple[ReportType, bytes, int]:
        """
        Parse a raw QMI SSC Control Message Input from a client.

        Parameters
        ----------
        message : bytes
            Raw QMI message to parse.

        Returns
        -------
        report_type : ReportType
            Report type requested by client.
        data : bytes
            Protobuf data passed by client.
        transaction_id : int
            Transaction ID to use for replies.

        Raises
        ------
        NotImplementedError
            If the QMI report type is not supported.
        """
        # QMI header validation and transaction ID
        message_type, transaction_id, message_id, message_length = QMI.parse_header(message)
        assert message_type == MessageType.REQUEST
        assert message_id == SSC_CONTROL_MSG_ID

        # TLV report type
        tlv_type, tlv_length, tlv_value = QMI.parse_tlv(message,
                                                        SSC_CONTROL_MSG_REPORT_TYPE_ID,
                                                        ValueType.INT)
        if tlv_value == ReportType.SMALL.value:
            report_type = ReportType.SMALL
        elif tlv_value == ReportType.LARGE.value:
            report_type = ReportType.LARGE
        else:
            raise NotImplementedError(f'Unknown report type: {tlv_value}')

        # TLV data
        tlv_type, tlv_length, tlv_value = QMI.parse_tlv(message,
                                                        SSC_CONTROL_MSG_DATA_ID,
                                                        ValueType.ARRAY)
        data = tlv_value

        return report_type, data, transaction_id

    @staticmethod
    def parse_protobuf_client_request(message: bytes) -> Tuple[int, int, int]:
        """
        Parse a Protobuf client request to extract sensor UID.

        Parameters
        ----------
        message : bytes
            Protobuf message to parse.

        Returns
        -------
        message_id : int
            ID of the Protobuf message
        uid_high : int
            UID 64 bits high.
        uid_low : int
            UID 64 bits low.
        """

        client_request = SscCommon.SscClientRequest()
        client_request.ParseFromString(message)
        uid_high = client_request.uid.high
        uid_low = client_request.uid.low
        message_id = client_request.msg_id

        return message_id, uid_high, uid_low

    @staticmethod
    def parse_protobuf_discovery_request(message: bytes) -> str:
        """
        Parse a Protobuf discovery request to extract the sensor data type to discover.

        Parameters
        ----------
        message : bytes
            Protobuf message to parse.

        Returns
        -------
        data_type : str
            Sensor data type.
        """

        client_request = SscCommon.SscClientRequest()
        client_request.ParseFromString(message)

        suid_request = SensorSuid.SscSuidRequest()
        suid_request.ParseFromString(client_request.request.msg)
        data_type = suid_request.data_type

        return data_type


    @staticmethod
    def generate_message_control_output(transaction_id: int) -> bytes:
        """
        Generate a QMI SSC Control Message Output as a reply.

        Parameters
        ----------
        transaction_id : int
            QMI transaction ID to use for replies.

        Returns
        -------
        message : bytes
            Generated QMI message as raw bytes.
        """
        message = QMI.generate_header(MessageType.RESPONSE,
                                      transaction_id,
                                      SSC_CONTROL_MSG_ID)

        # QMI Success response
        value = bytes()
        value += int(0x000).to_bytes(2, byteorder='little')
        value += int(0x000).to_bytes(2, byteorder='little')
        message = QMI.generate_tlv(message, SSC_CONTROL_MSG_SUCCESS_ID, value, ValueType.SEQUENCE)

        # SSC Client ID: not used by libssc
        value = bytes()
        value += int(SSC_CLIENT_ID).to_bytes(4, byteorder='little')
        message = QMI.generate_tlv(message, SSC_CONTROL_MSG_CLIENT_ID, value, ValueType.INT)

        # SSC Response (TODO): not used by libssc

        return message

    @staticmethod
    def generate_protobuf_discovery_response(data_type: str) -> bytes:
        """
        Generates a Protobuf message to discovery a sensor for a given data type.

        Parameters
        ----------
        data_type : str
            Sensor to discover for data type.

        Returns
        -------
        protobuf : bytes
            Protobuf message in bytes.

        Raises
        ------
        NotImplementedError
            If no sensor implementation can be found for the data type.
        """
        # Extract sensor UID by data type
        data = SSC._read_data()
        uid_high = None
        uid_low = None
        supported = False

        for entry in data:
            sensor = entry['sensor']
            if sensor['data_type'] == data_type:
                uid_high = int(sensor['uid_high'], 16)
                uid_low = int(sensor['uid_low'], 16)
                supported = True
                break

        # Generate discovery response 
        suid_response = SensorSuid.SscSuidResponse()
        suid_response.data_type = data_type
        if supported and uid_high is not None and uid_low is not None:
            sensor_uid = SscCommon.SscUid()
            sensor_uid.high = uid_high
            sensor_uid.low = uid_low
            suid_response.uid.append(sensor_uid)

        # Protobuf envelope to pack discovery message
        client_response_body = SscCommon.SscClientResponseBody()
        client_response_body.msg = suid_response.SerializeToString()
        client_response_body.msg_id = SSC_PROTOBUF_DISCOVERY_MSG_ID
        client_response_body.timestamp = int(time.time())

        client_response = SscCommon.SscClientResponse()
        client_response.response.append(client_response_body)
        client_response.uid.high = SSC_SUID_SENSOR_UID_HIGH 
        client_response.uid.low = SSC_SUID_SENSOR_UID_LOW

        protobuf = client_response.SerializeToString()
        return protobuf

    @staticmethod
    def generate_protobuf_attributes_response(uid_high: int, uid_low: int) -> bytes:
        """
        Generate a Protobuf message with all attributes of a sensor.

        Parameters
        ----------
        uid_high : int
            UID 64 bits high.
        uid_low : int
            UID 64 bits low.

        Returns
        -------
        protobuf : bytes
            Protobuf message in bytes.

        Raises
        ------
        NotImplementedError
            If no attributes can be found for the sensor UID.
        """
        # Extract sensor attributes for requested sensor by UID
        data = SSC._read_data()
        attributes = []

        for entry in data:
            sensor = entry['sensor']
            if int(sensor['uid_high'], 16) == uid_high and int(sensor['uid_low'], 16) == uid_low:
                for a in entry['attributes']:
                    attr_id = entry['attributes'][a]['id']
                    attr_value = entry['attributes'][a]['value']
                    if type(attr_value) is int:
                        attributes.append((attr_id, SscCommon.SscAttrValue(i=attr_value)))
                    elif type(attr_value) is float:
                        attributes.append((attr_id, SscCommon.SscAttrValue(f=attr_value)))
                    elif type(attr_value) is str:
                        attributes.append((attr_id, SscCommon.SscAttrValue(s=attr_value)))
                    elif type(attr_value) is bool:
                        attributes.append((attr_id, SscCommon.SscAttrValue(b=attr_value)))
                    elif type(attr_value) is list:
                        values = []
                        # Attributes with lists as type need to be converted to a list of ProtoBuf properties
                        for v in attr_value:
                            if type(v) is int:
                                values.append(SscCommon.SscAttrValue(i=v))
                            elif type(v) is float:
                                values.append(SscCommon.SscAttrValue(f=v))
                            elif type(v) is str:
                                values.append(SscCommon.SscAttrValue(s=v))
                            elif type(v) is bool:
                                values.append(SscCommon.SscAttrValue(b=v))
                            else:
                                raise NotImplementedError(f'Attribute value type in list unsupported: {v} ({type(v)})')
                        attributes.append((attr_id, values))
                    else:
                        raise NotImplementedError(f'Attribute value type unsupported: {attr_value} ({type(attr_value)})')
                break

        if not attributes:
            raise NotImplementedError(f'Unable to match sensor {uid_high},{uid_low} with attributes')

        # Generate attributes response 
        attr_response = SscCommon.SscAttrResponse()
        for a in attributes:
            attr = SscCommon.SscAttr()
            attr_array_value = SscCommon.SscAttrArrayValue()
            attr.id = a[0]

            # Attributes with a list need to be repeated
            if type(a[1]) is list:
                for element in a[1]:
                    attr.value_array.v.append(element)
            # Regular attribute
            else:
                attr.value_array.v.append(a[1])

            attr_response.attr.append(attr)

        # Protobuf envelope to pack attributes message
        client_response_body = SscCommon.SscClientResponseBody()
        client_response_body.msg = attr_response.SerializeToString()
        client_response_body.msg_id = SSC_PROTOBUF_RESPONSE_GET_ATTRIBUTES_MSG_ID
        client_response_body.timestamp = int(time.time())

        client_response = SscCommon.SscClientResponse()
        client_response.response.append(client_response_body)
        client_response.uid.high = uid_high 
        client_response.uid.low = uid_low

        protobuf = client_response.SerializeToString()
        return protobuf

    @staticmethod
    def generate_protobuf_sensor_measurement(uid_high: int, uid_low: int, index: int) -> bytes:
        """
        Instantiates a Protobuf sensor measurement message.

        Parameters
        ----------
        uid_high : int
            UID 64 bits high.
        uid_low : int
            UID 64 bits low.

        Returns
        -------
        protobuf : bytes
            Protobuf message in bytes.

        Raises
        ------
        NotImplementedError
            If no sensor measurements can be found for the sensor UID.
        """
        # Extract sensor measurements for requested sensor by UID
        data = SSC._read_data()
        data_type = ''
        measurements = []

        for entry in data:
            sensor = entry['sensor']
            data_type = sensor['data_type']
            if int(sensor['uid_high'], 16) == uid_high and int(sensor['uid_low'], 16) == uid_low:
                while len(entry['measurements']) <= index:
                    index = index - len(entry['measurements'])
                measurement = entry['measurements'][index]
                break

        if not measurement:
            raise NotImplementedError('Unable to match sensor {uid_high},{uid_low} with measurements')

        # Generate measurement message based on sensor data type
        if data_type == DataType.PROXIMITY.value:
            measurement_response = SensorProximity.SscProximityResponse()
            measurement_response.near = measurement['near']
            measurement_response.distance = measurement['distance']
        elif data_type == DataType.ACCELEROMETER.value:
            measurement_response = SensorAccelerometer.SscAccelerometerResponse()
            for m in measurement['acceleration']:
                measurement_response.acceleration.append(m)
        elif data_type == DataType.MAGNETOMETER.value:
            measurement_response = SensorMagnetometer.SscMagnetometerResponse()
            for m in measurement['magnetic_field']:
                measurement_response.magnetic_field.append(m)
        elif data_type == DataType.LIGHT.value:
            measurement_response = SensorLight.SscLightResponse()
            for m in measurement['intensity']:
                measurement_response.intensity.append(m)
        elif data_type == DataType.ROTATIONVECTOR.value:
            measurement_response = SensorCompass.SscRotationvectorResponse()
            for m in measurement['rotation']:
                measurement_response.rotation.append(m)
        else:
            raise NotImplementedError(f'Unsupported sensor {uid_high},{uid_low} for measurement generation')
        
        measurement_response.accuracy = measurement['accuracy']

        # Protobuf envelope to pack measurement message
        client_response_body = SscCommon.SscClientResponseBody()
        client_response_body.msg = measurement_response.SerializeToString()
        client_response_body.msg_id = measurement['id']
        client_response_body.timestamp = int(time.time())

        client_response = SscCommon.SscClientResponse()
        client_response.response.append(client_response_body)
        client_response.uid.high = uid_high 
        client_response.uid.low = uid_low

        protobuf = client_response.SerializeToString()
        return protobuf

    @staticmethod
    def generate_report_large_indication(data: bytes, transaction_id: int) -> bytes:
        """
        Instantiates a QMI SSC Report Large as a QMI indication.

        Parameters
        ----------
        data : bytes
            Protobuf binary data to pass along.
        transaction_id : int
            QMI transaction ID to use for replies.

        Returns
        -------
        msg : bytes
            The indication message as bytes.
        """
        message = QMI.generate_header(MessageType.INDICATION,
                                      transaction_id,
                                      SSC_REPORT_LARGE_MSG_ID)

        # SSC Client ID
        value = bytes()
        value += int(SSC_CLIENT_ID).to_bytes(8, byteorder='little')
        message = QMI.generate_tlv(message, SSC_REPORT_LARGE_CLIENT_ID, value, ValueType.INT)

        # SSC Data
        message = QMI.generate_tlv(message, SSC_REPORT_LARGE_DATA, data, ValueType.ARRAY)

        return message

    @staticmethod
    def generate_report_small_indication(data: bytes, transaction_id: int) -> bytes:
        """
        Instantiates a QMI SSC Small Large as a QMI indication.

        Parameters
        ----------
        data : bytes
            Protobuf binary data to pass along.
        transaction_id : int
            QMI transaction ID to use for replies.

        Returns
        -------
        msg : bytes
            The indication message as bytes.
        """
        message = QMI.generate_header(MessageType.INDICATION,
                                      transaction_id,
                                      SSC_REPORT_SMALL_MSG_ID)

        # SSC Client ID
        value = bytes()
        value += int(SSC_CLIENT_ID).to_bytes(4, byteorder='little')
        message = QMI.generate_tlv(message, SSC_REPORT_SMALL_CLIENT_ID, value, ValueType.INT)

        # SSC Data
        message = QMI.generate_tlv(message, SSC_REPORT_SMALL_DATA, data, ValueType.ARRAY)

        return message
